/******************************************************************************
Asservissement de suivi de ligne d'un robot mobile.
Ce programme est destiné à fonctionner avec Mobilos, 
disponible à l'adresse suivante:
http://boutique.3sigma.fr/50-mobilos.html


Auteur: 3Sigma
Version 2.2 - 29/04/2014
*******************************************************************************/

// Inclusion d'une bibliothèque permettant l'exécution à cadence fixe
// d'une partie du programme. Télécharger à l'adresse http://www.3sigma.fr/telechargements/FlexiTimer2.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
// Pour plus de détails, voir les pages (en anglais): 
// http://www.arduino.cc/playground/Main/FlexiTimer2
// https://github.com/wimleers/flexitimer2
#include <FlexiTimer2.h>

// Inclusion d'une bibliothèque permettant de lire et d'écrire plus rapidement sur les entrées-sorties digitales.
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/digitalWriteFast.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <digitalWriteFast.h> 

// Inclusion d'une bibliothèque de gestions des capteurs de suivi de ligne
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/QTRSensors.zip
#include <QTRSensors.h>

// Définitions et déclarations pour le codeur incrémental du moteur droit
#define codeurDroitInterruptionA 0
#define codeurDroitPinA 2
#define codeurDroitPinB 8
#define Nmoy 10
volatile long ticksCodeurDroit = 0;
static int indiceTicksCodeurDroit = 0;
static int ticksCodeurDroitTab[Nmoy];
static int codeurDroitDeltaPos;

// Définitions et déclarations pour le codeur incrémental du moteur gauche
#define codeurGaucheInterruptionA 1
#define codeurGauchePinA 3
#define codeurGauchePinB 9
volatile long ticksCodeurGauche = 0;
static int indiceTicksCodeurGauche = 0;
static int ticksCodeurGaucheTab[Nmoy];
static int codeurGaucheDeltaPos;

// Définitions et déclarations pour le moteur à courant continu.
#define directionMoteurDroit  4
#define pwmMoteurDroit  5
#define directionMoteurGauche  7
#define pwmMoteurGauche  6

// Déclarations pour les capteurs de suivi de ligne
QTRSensorsAnalog qtr((unsigned char[]) {0, 1, 2}, 3);
unsigned int sensors[3];

// Ce programme envoie par liaison série des données à l'ordinateur connecté à la carte.
// Cadence d'envoi des données en ms
#define TSDATA 100
unsigned long tempsDernierEnvoi = 0;
unsigned long tempsCourant = 0;

// Certaines parties du programme sont exécutées à cadence fixe grâce à la bibliothèque FlexiTimer2.
// Cadence d'échantillonnage en ms
#define CADENCE_MS 10
volatile double dt = CADENCE_MS/1000.;
volatile double temps = -CADENCE_MS/1000.;

// Le gyropode est asservi grâce à plusieurs régulateurs de type PI
// On déclare ci-dessous les variables et paramètres nécessaires à l'asservissement et au régulateur
static double R = 0.045; // Rayon d'une roue
static double W = 0.14; // Largeur du gyropode
static double vxmes = 0.; // vitesse longitudinale mesurée
static double xidotmes = 0.; // vitesse de rotation mesurée
static double Kpvx = 5.; // gain proportionnel pour l'asservissement de vitesse longitudinale
static double Kivx = 45.; // gain intégral pour l'asservissement de vitesse longitudinale
static double commande_vx = 0.; // commande pour l'asservissement de vitesse longitudinale
static double P_vx = 0.; // action proportionnelle pour l'asservissement de vitesse longitudinale
static double I_vx = 0.; // action intégrale pour l'asservissement de vitesse longitudinale

// Déclarations pour la commande des moteurs
static double omegaDroit = 0.;
static double omegaGauche = 0.;
static double commandeDroit = 0.;
static double commandeGauche = 0.;

// Déclarations pour les consignes de mouvement
static double vxref = 0.;
static double Tcal = 1.5;

// Déclarations pour la tension batterie
static double tensionBatterie;

// Suivi de ligne
// Déclarations pour le suivi de ligne
static double erreurPrecedente = 0.;
static double erreurSuiviLigne = 0.;
static int position = 0;

// Déclarations pour le PD de suivi de ligne
#define VXINIT 0.1 // vitesse initiale (m/s)
#define TSSUIVI 100 // période de calcul de l'asservissement de suivi de ligne (en ms)
unsigned long tempsDernierCalcul = 0;
unsigned long tempsCourantSuivi = 0;
static double commande_suivi = 0.; // commande de vitesse angulaire calculée par le PD
static double commande_suivi_avant_sat = 0.; // valeur de la commande avant la saturation (voir ci-dessous)
static double commande_suivi_max = 3.; // valeur max de la tension de commande
static double commande_suivi_min = -3.; // valeur min (ou max en négatif) de la tension de commande
static double Kp = 5.; // gain proportionnel
static double Kd = 20.; // gain dérivé
static double Tf = 0.2; // constante de temps de filtrage de l'action dérivée du PD
static double P_x = 0.; // valeur de l'action proportionnelle
static double D_x = 0.; // valeur de l'action dérivée
static double D_x_temp = 0.; // pour calculs intermédiaires
static double deltaTemps = 0.1;
// Variables intermédiaires
static double ad = 0;
static double bd = 0;


// Initialisations
void setup(void) {
  
  // Codeur incrémental moteur droit
  pinMode(codeurDroitPinB, INPUT);      // entrée digitale pin B codeur
  digitalWrite(codeurDroitPinB, HIGH);  // activation de la résistance de pullup
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurDroitPinA (définie à la fin du programme)
  attachInterrupt(codeurDroitInterruptionA, GestionInterruptionCodeurDroitPinA, CHANGE);

  // Codeur incrémental moteur gauche
  pinMode(codeurGauchePinB, INPUT);      // entrée digitale pin B codeur
  digitalWrite(codeurGauchePinB, HIGH);  // activation de la résistance de pullup
  // A chaque changement de niveau de tension sur le pin A du codeur,
  // on exécute la fonction GestionInterruptionCodeurGauchePinA (définie à la fin du programme)
  attachInterrupt(codeurGaucheInterruptionA, GestionInterruptionCodeurGauchePinA, CHANGE);
 
  // Moteur à courant continu droit
  pinMode(directionMoteurDroit, OUTPUT);
  pinMode(pwmMoteurDroit, OUTPUT);

  // Moteur à courant continu gauche
  pinMode(directionMoteurGauche, OUTPUT);
  pinMode(pwmMoteurGauche, OUTPUT);
  
  // Suivi de ligne
  pinMode(13, OUTPUT);

  // Liaison série.
  Serial.begin(57600);
  Serial.flush();
  
  // Compteur d'impulsions des encodeurs
  ticksCodeurDroit = 0;
  ticksCodeurGauche = 0;  
  
  /* Le programme principal est constitué:
     - de la traditionnelle boucle "loop" des programmes Arduino, dans laquelle on lit 
       en permanence la liaison série pour récupérer les nouvelles consignes de mouvement
     - de la fonction "isrt", dont l'exécution est définie à cadence fixe par les deux instructions
       ci-dessous. La bonne méthode pour exécuter une fonction à cadence fixe, c'est de l'exécuter sur interruption
       (la boucle "loop" est alors interrompue régulièrement, à la cadence voulue, par un timer).
       Une mauvaise méthode serait d'utiliser la fonction Arduino "delay". En effet, la fonction "delay" définit
       un délai entre la fin des instructions qui la précèdent et le début des instructions qui la suivent, mais si
       on ne connait pas le temps d'exécution de ces instructions (ou si celui-ci est inconnu), on n'obtiendra
       jamais une exécution à cadence fixe.
       Il est nécessaire d'exécuter la fonction "isrt" à cadence fixe car cette fonction:
       - doit renvoyer des données à cadence fixe sur la liaison série, pour monitoring
       - calcule l'asservissement de verticalité et de mouvement, conçu pour fonctionner à une cadence bien spécifiée
     
  */
  FlexiTimer2::set(CADENCE_MS, 1/1000., isrt); // résolution timer = 1 ms
  FlexiTimer2::start();

}


// Boucle principale
void loop() {


}

// Fonction excutée sur interruption
void isrt(){

  // Compteur de boucle et variable d'activation de l'asservissement
  int i;

  // Mesure de la vitesse des moteurs grâce aux codeurs incrémentaux
  
  // On calcule une moyenne glissante de la vitesse sur les Nmoy derniers échantillons
  // Nombre de ticks codeur depuis la dernière fois accumulé pour les 1à derniers échantillons
  // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurDroitPinA et GestionInterruptionCodeurDroitPinA,
  // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
  for (int i=0; i<Nmoy; i++) {
    ticksCodeurDroitTab[i] += ticksCodeurDroit;
  }  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeurDroit = 0;
  
  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurDroitDeltaPos = ticksCodeurDroitTab[indiceTicksCodeurDroit];
  // Remise à zéro du compteur d'impulsion codeur pour l'échantillon courant
  ticksCodeurDroitTab[indiceTicksCodeurDroit] = 0;
  
  // Mise à jour de l'indice d'échantillon courant
  indiceTicksCodeurDroit++;
  if (indiceTicksCodeurDroit==Nmoy) {
    indiceTicksCodeurDroit = 0;
  }
  
  // On calcule une moyenne glissante de la vitesse sur les Nmoy derniers échantillons
  // Nombre de ticks codeur depuis la dernière fois accumulé pour les 1à derniers échantillons
  // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurGauchePinA et GestionInterruptionCodeurGauchePinA,
  // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
  for (int i=0; i<Nmoy; i++) {
    ticksCodeurGaucheTab[i] += ticksCodeurGauche;
  }  
  // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
  ticksCodeurGauche = 0;
  
  // Pour l'échantillon courant, calcule de l'angle de rotation du moteur pendant la période d'échantillonnage
  codeurGaucheDeltaPos = ticksCodeurGaucheTab[indiceTicksCodeurGauche];
  // Remise à zéro du compteur d'impulsion codeur pour l'échantillon courant
  ticksCodeurGaucheTab[indiceTicksCodeurGauche] = 0;
  
  // Mise à jour de l'indice d'échantillon courant
  indiceTicksCodeurGauche++;
  if (indiceTicksCodeurGauche==Nmoy) {
    indiceTicksCodeurGauche = 0;
  }

  // Calcul de la vitesse de rotation. C'est le nombre d'impulsions converti en radian, divisé par la période d'échantillonnage
  // On fait un calcul glissant sur Nmoy échantillons, d'où le Nmoy*dt
  // ATTENTION: le facteur 2 devant est là car avec 2 moteurs, la résolution est divisée par 2
  omegaDroit = -2*((2.*3.141592*((double)codeurDroitDeltaPos))/1632.)/(Nmoy*dt);  // en rad/s
  omegaGauche = 2*((2.*3.141592*((double)codeurGaucheDeltaPos))/1632.)/(Nmoy*dt);  // en rad/s
 
  
  // Définition des entrées de la fonction d'asservissement
  vxmes = (omegaDroit + omegaGauche)*R/2;
  xidotmes = (omegaDroit - omegaGauche)*R/W;

  /******* Calcul du PI sur vx *******/
  
  // Terme proportionnel (la transformation de la commande par retour d'état en PI
  // conduit à une référence nulle, d'où le 0.*vxref)
  P_vx = Kpvx * (0.*vxref - vxmes);
  
  // Calcul de la commande avant saturation
  commande_vx = P_vx + I_vx;
  
  
  // Terme intégral (sera utilisé lors du pas d'échantillonnage suivant)
  I_vx = I_vx + Kivx * dt * (vxref - vxmes);

  /******* Fin Calcul du PI sur vx *******/
    
  /********************* Suivi de ligne **************************/
  /***************************************************************/
  vxref = VXINIT;
      
  
  // Calibration
  if (temps < Tcal) {
    qtr.calibrate();
  }
  
  else {
  
    digitalWrite(13,HIGH);
    
    // Le calcul de l'asservissement de suivi de ligne est sous-échantillonné
    // par rapport à l'interruption de base (par défaut, exécution une fois sur dix,
    // c'est-à-dire à 10 Hz)
    tempsCourantSuivi = millis();
    if (tempsCourantSuivi-tempsDernierCalcul > TSSUIVI) {
      
      position = qtr.readLine(sensors);
      erreurSuiviLigne = 1. - position/1000.;
              
      /******* Calcul du PD de suivi de ligne *******/
      // Temps entre deux mesures
      deltaTemps = dt*TSSUIVI/CADENCE_MS;
      
      // Paramètre de filtrage de l'action dérivée
      ad = Tf/(Tf+deltaTemps);
      bd = Kd/(Tf+deltaTemps);
      
      // Calcul du terme dérivé
      D_x_temp = erreurSuiviLigne - erreurPrecedente;
      D_x = ad * D_x + bd * D_x_temp;
      
            
      // Calcul du terme proportionnel
      P_x = Kp * erreurSuiviLigne;
          
      // Calcul de la commande avant saturation
      commande_suivi_avant_sat = P_x + D_x;
      
      // Application de la saturation sur la commande
      if (commande_suivi_avant_sat > commande_suivi_max) {
        commande_suivi = commande_suivi_max;
      }
      else if (commande_suivi_avant_sat < commande_suivi_min) {
        commande_suivi = commande_suivi_min;
      }
      else {
        commande_suivi = commande_suivi_avant_sat;
      }
      
      // Détermination des valeurs qui seront les valeurs précédentes du prochain calcul
      erreurPrecedente = erreurSuiviLigne;
      tempsDernierCalcul = tempsCourantSuivi;
      /******* Fin Calcul du PD de suivi de ligne *******/
    }
  }
  
  /***************** Fin suivi de ligne **************************/
  /***************************************************************/

  
  // Calcul des commandes des moteurs
  commandeDroit = (commande_vx + commande_suivi);
  commandeGauche = (commande_vx - commande_suivi);
  
  // Application des commandes aux moteurs
  tensionBatterie = 7.4;
  CommandeMoteurDroit(-commandeDroit, tensionBatterie);
  CommandeMoteurGauche(-commandeGauche, tensionBatterie);
  
  // Ecriture des données sur la liaison série pour monitoring
  ecritureData();
  
  // Incrémentation du temps courant
  temps += dt;

}


void ecritureData(void) {
  // Ecriture des données en sortie tous les TSDATA millisecondes
  
  // On envoie les données s'il s'est écoulé plus de TSDATA secondes
  // entre l'instant courant et le dernier envoi
  tempsCourant = millis();
  if (tempsCourant-tempsDernierEnvoi > TSDATA) {
    Serial.print(temps);
    
    Serial.print(",");
    Serial.print(position);
    Serial.print(",");
    Serial.print(erreurSuiviLigne);
    Serial.print(",");
    Serial.print(commande_suivi);
    Serial.print(",");
    Serial.print(sensors[0]);
    Serial.print(",");
    Serial.print(sensors[1]);
    Serial.print(",");
    Serial.print(sensors[2]);
  
    Serial.print("\r");
    Serial.print("\n");
    
    tempsDernierEnvoi = tempsCourant;
  }  
}

void CommandeMoteurDroit(double commande, double tensionBatterie) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;

  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commandeDroit(commande);

  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(255*(tension/tensionBatterie));

  // Saturation par sécurité
  if (tension_int>255) {
    tension_int = 255;
  }

  if (tension_int<-255) {
    tension_int = -255;
  }

  // Commande PWM
  if (tension_int>=0) {
    digitalWrite(directionMoteurDroit, LOW);
    analogWrite(pwmMoteurDroit, tension_int);
  }

  if (tension_int<0) {
    digitalWrite(directionMoteurDroit, HIGH);
    analogWrite(pwmMoteurDroit, -tension_int);
  }
}


double redresse_commandeDroit(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.5); 
  }
  else if (commande<0) {
    return(commande-0.5); 
  }
  else {
    return(0); 
  }
}


void CommandeMoteurGauche(double commande, double tensionBatterie) {
  // Cette fonction calcule et envoi les signaux PWM au pont en H
  // en fonction des tensions de commande et d'alimentation
  int tension_int;
  double tension;

  // L'ensemble pont en H + moteur n'est pas linéaire
  // On redresse la caractéristique avec cette fonction
  tension = redresse_commandeGauche(commande);

  // Normalisation de la tension d'alimentation par
  // rapport à la tension d'alimentation
  tension_int = (int)(255*(tension/tensionBatterie));

  // Saturation par sécurité
  if (tension_int>255) {
    tension_int = 255;
  }
  if (tension_int<-255) {
    tension_int = -255;
  }

  // Commande PWM
  if (tension_int>=0) {
    digitalWrite(directionMoteurGauche, LOW);
    analogWrite(pwmMoteurGauche, tension_int);
  }
  if (tension_int<0) {
    digitalWrite(directionMoteurGauche, HIGH);
    analogWrite(pwmMoteurGauche, -tension_int);
  }
}


double redresse_commandeGauche(double commande) {
  // Cette fonction redresse la non-linéarité de l'ensemble pont en H + moteur
  if (commande>0) {
    return(commande+0.5); 
  }
  else if (commande<0) {
    return(commande-0.5); 
  }
  else {
    return(0); 
  }
}

 
void GestionInterruptionCodeurDroitPinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental droit
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurDroitPinA) == digitalReadFast2(codeurDroitPinB)) {
    ticksCodeurDroit++;
  }
  else {
    ticksCodeurDroit--;
  }
}


void GestionInterruptionCodeurGauchePinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental gauche
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurGauchePinA) == digitalReadFast2(codeurGauchePinB)) {
    ticksCodeurGauche++;
  }
  else {
    ticksCodeurGauche--;
  }
}

